using System;
using System.Collections;
using System.Web;

namespace PageBoxGrid
{
	/// <summary>
	/// <p>Part of the Coordinator set. Event management.</p>
	/// <p>Copyright (c) 2002 Alexis Grandemange<br/>
	/// Mail: alexis.grandemange@pagebox.net</p>
	/// <pre>This program is free software; you can redistribute it and/or
	/// modify it under the terms of the GNU Lesser General Public
	/// License as published by the Free Software Foundation; version
	/// 2.1 of the License.
	/// This library is distributed in the hope that it will be useful,
	/// but WITHOUT ANY WARRANTY; without even the implied warranty of
	/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	/// GNU Lesser General Public License for more details.
	/// A copy of the GNU Lesser General Public License lesser.txt should be
	/// included in the distribution.</pre>
	/// </summary>
	public class CallbackCoordinator
	{
		/// <summary>
		/// <p>Key: Web server session ID</p>
		/// <p>Value: Hashtable<ul>
		/// <li>Key: ID of the Event source</li>
		/// <li>Value: Hashtable<ul>
		/// <li>Key: ID of the Event destination</li>
		/// <li>Value: Callback object</li>
		/// </ul></li>
		/// </ul>
		/// </p>
		/// </summary>
		private static Hashtable sessionEvents = new Hashtable();

		/// <summary>
		/// Called by an event destination to subscribe for an event
		/// </summary>
		/// <param name="sourceID">ID of the Event source</param>
		/// <param name="sessionID">Web server session ID</param>
		/// <param name="ID">ID of the Event destination</param>
		/// <param name="cb">Object provided by the subscriber</param>
		public static void Subscribe(string sourceID, string sessionID, string ID, Callback cb) 
		{
			lock(typeof(CallbackCoordinator)) 
			{
				Hashtable sourceEvents, destEvents;
				bool newDest = false, newSource = false;
				if (!sessionEvents.Contains(sessionID)) 
				{
					sourceEvents = new Hashtable();
					newSource = true;
				}
				else
					sourceEvents = (Hashtable)sessionEvents[sessionID];
				if (!sourceEvents.Contains(sourceID)) 
				{
					destEvents = new Hashtable();
					newDest = true;
				}
				else
					destEvents = (Hashtable)sourceEvents[sourceID];
				if (!destEvents.Contains(ID))
					destEvents.Add(ID, cb);
				else
					destEvents[ID] = cb;
				if (newDest)
					sourceEvents.Add(sourceID, destEvents);
				if (newSource)
					sessionEvents.Add(sessionID, sourceEvents);
			}
		}

		/// <summary>
		/// Called by the event source to notify an event.
		/// Notes:<ol>
		/// <li>If there is no subscriber the answer is true</li>
		/// <li>An exception = a false answer</li>
		/// <li>All subscribers are notified even if a subscriber answers false</li>
		/// </ol>
		/// </summary>
		/// <param name="sourceID">ID of the Event source</param>
		/// <param name="sessionID">Web server session ID</param>
		/// <param name="req">Object to pass to the Notify method of the destination</param>
		/// <param name="resp">Hashtable of responses whose key is the destination ID</param>
		/// <param name="request">HTTP request object of the source</param>
		/// <returns>True if Notify methods of all subscribers answered true</returns>
		public static bool Notify(string sourceID, string sessionID, object req, ref Hashtable resp,
			HttpRequest request) 
		{
			lock(typeof(CallbackCoordinator)) 
			{
				if (!sessionEvents.Contains(sessionID))
					return true;
				Hashtable sourceEvents = (Hashtable)sessionEvents[sessionID];
				if (!sourceEvents.Contains(sourceID))
					return true;
				Hashtable destEvents = (Hashtable)sourceEvents[sourceID];
				IDictionaryEnumerator ide = destEvents.GetEnumerator();
				bool rc = true;
				while(ide.MoveNext()) 
				{
					Callback cb = (Callback)ide.Value;
					try 
					{
						object r = null;
						if (!cb.Notify(sourceID, req, ref r, request))
							rc = false;
						if ((r != null) && (resp != null))
							resp[ide.Key] = r;
					}
					catch(Exception e) 
					{
						if (resp != null)
							resp[ide.Key] = e;
						rc = false;
					}
				}
				return rc;
			}
		}
	}

	/// <summary>
	/// Interface provided by the subscriber
	/// </summary>
	public interface Callback 
	{
		/// <summary>
		/// Notify an event to a subscriber
		/// </summary>
		/// <param name="sourceID">ID of the Event source</param>
		/// <param name="req">Object provided by the Event source when it invokes Notify</param>
		/// <param name="resp">Object returned by the Event target</param>
		/// <param name="request">HTTP request object of the source</param>
		/// <returns>Return false to indicate a failure</returns>
		bool Notify(string sourceID, object req, ref object resp, HttpRequest request); 
	}
}
