using System;
using System.Web.Services.Protocols;
using System.Web.Services;
using System.Collections;
using System.Web.Mail;
using System.IO;
using System.Text;
using System.Threading;
using System.Net;
using System.Net.Sockets;
using System.Runtime.Serialization.Formatters.Binary;

namespace PageBoxGrid
{
	/// <summary>
	/// Transport modes supported by PageBox Grid
	/// </summary>
	public enum Transport 
	{
		UDP,
		SMTP,
		Memory
	}

	/// <summary>
	/// <p>Grid API.
	/// See <a href="http://pagebox.net/grid.html">Grid</a> for explanations.</p>
	/// <p>Copyright (c) 2002 Alexis Grandemange<br/>
	/// Mail: alexis.grandemange@pagebox.net</p>
	/// <pre>This program is free software; you can redistribute it and/or
	/// modify it under the terms of the GNU Lesser General Public
	/// License as published by the Free Software Foundation; version
	/// 2.1 of the License.
	/// This library is distributed in the hope that it will be useful,
	/// but WITHOUT ANY WARRANTY; without even the implied warranty of
	/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	/// GNU Lesser General Public License for more details.
	/// A copy of the GNU Lesser General Public License lesser.txt should be
	/// included in the distribution.</pre>
	/// </summary>
	public class Grid
	{
		/// <summary>
		/// GridEntry records
		/// </summary>
		public GridEntry[] destinations;
		/// <summary>
		/// SMTP server used to send messages.
		/// </summary>
		private string SMTPServer;
		/// <summary>
		/// Thread waiting for mails on POP server.
		/// </summary>
		private POPListener POPThread;
		/// <summary>
		/// Thread waiting for datagrams on UDP port.
		/// </summary>
		private UDPListener UDPThread;
		/// <summary>
		/// Thread waiting for datagrams on multicast port.
		/// </summary>
		private MulticastListener multicastThread;
		/// <summary>
		/// Receive timeout
		/// </summary>
		private TimeSpan timeout;
		/// <summary>
		/// GridEntry of this Grid instance.
		/// </summary>
		public GridEntry ge;
		/// <summary>
		/// Multicast address.
		/// </summary>
		private IPAddress multicastAddress;
		/// <summary>
		/// Multicast port.
		/// </summary>
		private int multicastPort;
		/// <summary>
		/// Time To Live (number of hops)
		/// </summary>
		private int ttl;
		/// <summary>
		/// Sending UDP client.
		/// </summary>
		private UdpClient uc;
		/// <summary>
		/// Sending multicast client.
		/// </summary>
		private UdpClient ucm;
		/// <summary>
		/// Multicast end point (address + port)
		/// </summary>
		private IPEndPoint multicastEP;
		/// <summary>
		/// Used in SMTPScatter.
		/// Key: mail address.
		/// Value: ArrayList of GridNb.
		/// </summary>
		public Hashtable mailDests = new Hashtable();
		/// <summary>
		/// Used in UDPScatter.
		/// Key: IPEndPoint.
		/// Value: ArrayList of GridNb.
		/// </summary>
		public Hashtable UDPDests = new Hashtable();
		/// <summary>
		/// Communicator name
		/// </summary>
		private string communicator;
		/// <summary>
		/// Log object used to write errors on a file.
		/// </summary>
		private Log log;
		/// <summary>
		/// Message handler.
		/// </summary>
		public GridCallback gc = null;

		/// <summary>
		/// Number of partners in Grid
		/// </summary>
		public int DestNumber 
		{
			get { return destinations.Length; }
		}

		/// <summary>
		/// 1. Registers on its PageBox
		/// 2. Retrieves its partners and build their list
		/// 3. Connects to listening threads
		/// Version 0.0.5d: parm parameter.
		/// </summary>
		/// <param name="PageBoxUrl">URL of the deploying PageBox</param>
		/// <param name="repositoryUrl">URL of the communicator repository</param>
		/// <param name="communicator">Communicator name</param>
		/// <param name="UDPAddress">UDP address</param>
		/// <param name="UDPPort">UDP port</param>
		/// <param name="multicastAddress">UDP multicast address:
		/// must be unique for all objects with the same communicator name</param>
		/// <param name="multicastPort">UDP multicast port</param>
		/// <param name="ttl">Time to live = number of hops (multicast)</param>
		/// <param name="SMTPServer">SMTP server used to send messages</param>
		/// <param name="MailAddress">Mail address of this instance</param>
		/// <param name="POPServer">IP address of the POP server</param>
		/// <param name="POPUser">POP user name</param>
		/// <param name="POPPassword">POP password</param>
		/// <param name="parm">Extra parameter</param>
		/// <param name="LogFile">File used to log Grid errors</param>
		/// <param name="timeout">Period between two POP3 read</param>
		public Grid(string PageBoxUrl, string repositoryUrl, string communicator, string UDPAddress,
			int UDPPort, string multicastAddress, int multicastPort, int ttl, string SMTPServer,
			string MailAddress, string POPServer, string POPUser, string POPPassword, string parm,
			string LogFile, double timeout)
		{
			log = new Log(LogFile);
			this.SMTPServer = SMTPServer;
			this.timeout = TimeSpan.FromSeconds((timeout * 60) + 1);
			string url = PageBoxUrl.Replace("Deploy.asmx", "ActiveNaming.asmx");
			ActiveProxy ap = new ActiveProxy(url);
			try 
			{
				ap.Discover();
			}
			catch(Exception e) 
			{
				log.write("Grid.Grid(" + PageBoxUrl + ", " + SMTPServer + ", " + MailAddress + 
					") returned " + url + " PageBox not found");
				return;
			}
			if (multicastAddress != null)
				this.multicastAddress = IPAddress.Parse(multicastAddress);
			else
				this.multicastAddress = null;
			this.multicastPort = multicastPort;
			this.ttl = ttl;
			this.communicator = communicator;
			if ((POPServer != null) && (POPUser != null) && (POPPassword != null))
				POPThread = POPListener.GetListener(POPServer, POPUser, POPPassword, log, 
					timeout, this);
			if ((UDPAddress != null) && (UDPPort != 0))
				UDPThread = UDPListener.GetListener(UDPAddress, UDPPort, log, this);
			if (UDPThread != null) 
			{
				if (multicastAddress != null)
					multicastThread = MulticastListener.GetListener(multicastAddress,
						multicastPort, ttl, log, this);
			}
			retrievePartners(ap, repositoryUrl, PageBoxUrl);
			register(ap, UDPAddress, UDPPort, MailAddress, url, parm);
			cacheUDPMail();
		}

		/// <summary>
		/// Populates mailDests and UDPDests (mail and UDP address strings).
		/// </summary>
		private void cacheUDPMail() 
		{
			ArrayList al = null;
			for (int i = 0; i < destinations.Length; ++i) 
			{
				if (destinations[i].mailAddress != null) 
				{
					if (!mailDests.Contains(destinations[i].mailAddress)) 
					{
						al = new ArrayList();
						al.Add(destinations[i].gridNb);
						mailDests.Add(destinations[i].mailAddress, al);
					} else {
						al = (ArrayList)mailDests[destinations[i].mailAddress];
						al.Add(destinations[i].gridNb);
					}
				}
				if ((destinations[i].UDPAddress != null) && (destinations[i].UDPPort != 0)) 
				{
					IPEndPoint iep = new IPEndPoint(destinations[i].addr,
						destinations[i].UDPPort);
					if (!UDPDests.Contains(iep)) 
					{
						al = new ArrayList();
						al.Add(destinations[i].gridNb);
						UDPDests.Add(iep, al);
					} 
					else 
					{
						al = (ArrayList)UDPDests[iep];
						al.Add(destinations[i].gridNb);
					}
				}
			}
		}
		
		/// <summary>
		/// Publish the process to the PageBox
		/// </summary>
		/// <param name="ap">Proxy on the ActiveNaming web service</param>
		/// <param name="UDPAddress">UDP address</param>
		/// <param name="UDPPort">UDP port</param>
		/// <param name="MailAddress">This process mail address</param>
		/// <param name="url">URL of the controlling ActiveNaming</param>
		/// <param name="parm">Extra parameter</param>
		private void register(ActiveProxy ap, string UDPAddress, int UDPPort, string MailAddress,
			string url, string parm) 
		{
			ge = GridEntry.getGridEntry(UDPAddress, UDPPort, MailAddress, url, parm, this);
			try 
			{
				ArrayList al = deserializeGEs(ap.GetKeyRange("PageBoxGrid.dll", communicator));
				al.Add(ge);
				ap.SetKeyRange("PageBoxGrid.dll", communicator, serializeGEs(al));
			}
			catch(System.Net.WebException we) 
			{
				log.write("Grid.register(" + communicator + ", " + UDPAddress + ", " +
					UDPPort + ", " + MailAddress + ", " + parm + ") returned " + we.Message);
			}
			catch(System.Web.HttpException he)
			{
				log.write("Grid.register(" + communicator + ", " + UDPAddress + ", " +
					UDPPort + ", " + MailAddress + ", " + parm + ") returned " + he.Message);
			}
			catch(SoapException se) 
			{
				log.write("Grid.register(" + communicator + ", " + UDPAddress + ", " +
					UDPPort + ", " + MailAddress + ", " + parm + ") returned " + se.Message);
			}
		}

		/// <summary>
		/// 1. Retrieve other processes of the process group
		/// 2. Notifies other processes in the process group
		/// </summary>
		/// <param name="ap">Proxy on the ActiveNaming web service</param>
		/// <param name="repositoryUrl">Communicator repository</param>
		/// <param name="PageBoxUrl">URL of the controlling PageBox</param>
		private void retrievePartners(ActiveProxy ap, string repositoryUrl, string PageBoxUrl) 
		{
			try 
			{
				Candidate[] candidates = ap.GetCandidate(repositoryUrl, "PageBoxGrid.dll", 
					communicator);
				ArrayList dests = new ArrayList();
				for (int i = 0; i < candidates.Length; ++i) 
				{
					ArrayList al = deserializeGEs(candidates[i].keyRange);
					IEnumerator ie = al.GetEnumerator();
					while(ie.MoveNext())
						dests.Add((GridEntry)ie.Current);
				}
				destinations = (GridEntry [])dests.ToArray(typeof(GridEntry));
			}
			catch(System.Net.WebException we) 
			{
				log.write("Grid.retrievePartners(" + repositoryUrl + ", " + communicator +
					") returned " + we.Message);
			}
			catch(System.Web.HttpException he)
			{
				log.write("Grid.retrievePartners(" + repositoryUrl + ", " + communicator +
					") returned " + he.Message);
			}
			catch(SoapException se) 
			{
				log.write("Grid.retrievePartners(" + repositoryUrl + ", " + communicator +
					") returned " + se.Message);
			}
			GridCommand gc = new GridCommand(ge, Command.Add);
			if (Scatter(gc, Transport.UDP) == -1)
				Scatter(gc, Transport.SMTP);
		}

		/// <summary>
		/// Send an object to another process
		/// </summary>
		/// <param name="o">Object to send</param>
		/// <param name="dest">Destination number</param>
		/// <param name="transport_mode">UDP or SMTP</param>
		/// <returns>-1 in case of failure</returns>
		public int Send(object o, int dest, Transport transport_mode) 
		{
			if ((dest < 0) || (dest > (destinations.Length - 1)))
				return -1;
			Grid destGrid = GridEntry.getGrid(destinations[dest]);
			if (destGrid != null)
				return MemorySend(o, destGrid, transport_mode);
			switch(transport_mode) 
			{
				case Transport.UDP:
					return UDPSend(o, dest);
				case Transport.SMTP:
					return SMTPSend(o, dest);
			}
			return -1;
		}

		/// <summary>
		/// Send an object to another thread in the same OS process
		/// </summary>
		/// <param name="o">Object to send</param>
		/// <param name="destGrid">Destination Grid</param>
		/// <param name="transport_mode">UDP or SMTP</param>
		/// <returns>-1 in case of error</returns>
		private int MemorySend(object o, Grid destGrid, Transport transport_mode) 
		{
			Message m = new Message();
			m.recv = o;
			m.transport_mode = Transport.Memory;
			m.from = -1;
			lock(destGrid) 
			{
				for (int i = 0; i < destGrid.destinations.Length; ++i) 
				{
					if (destGrid.destinations[i].gridNb != ge.gridNb)
						continue;
					if (transport_mode == Transport.SMTP) 
					{
						if (destGrid.destinations[i].mailAddress.Equals(ge.mailAddress))
						{
							m.from = i;
							break;
						} 
						else
							continue;
					}
					if (transport_mode == Transport.UDP) 
					{
						if (destGrid.destinations[i].UDPAddress.Equals(ge.UDPAddress) &&
							(destGrid.destinations[i].UDPPort == ge.UDPPort)) 
						{
							m.from = i;
							break;
						}
					}
				}
				if (destGrid.gc != null) 
					destGrid.gc.Notify(m.transport_mode, m.from, m.recv);
				else 
				{
					destGrid.clientQueue.Enqueue(m);
					if (receivePending) 
						Monitor.Pulse(destGrid);
				}
			}
			return 1;
		}

		/// <summary>
		/// Queue used for memory communication
		/// </summary>
		private Queue clientQueue = new Queue();
		/// <summary>
		/// Receive flag used in memory replication
		/// </summary>
		private bool receivePending = false;

		/// <summary>
		/// Send an object to another process in UDP.
		/// </summary>
		/// <param name="o">Object to send</param>
		/// <param name="dest">Destination number</param>
		/// <returns>-1 in case of error</returns>
		private int UDPSend(object o, int dest) 
		{
			if ((destinations[dest].UDPAddress == null) || (destinations[dest].UDPPort == 0))
				return -1;
			byte[] buf = UDPPrepare(o, dest);
			if (uc.Send(buf, buf.Length, destinations[dest].UDPAddress,
				destinations[dest].UDPPort) < 0)
				return -1;
			return 1;
		}

		/// <summary>
		/// Prepare the connection and serialize the object.
		/// </summary>
		/// <param name="o">Object to send</param>
		/// <param name="dest">Destination number</param>
		/// <returns>Serialized object or null in case of error</returns>
		private byte[] UDPPrepare(object o, int dest) 
		{
			if ((ge.addr == null) || (ge.UDPPort == 0))
				return null;
			if (uc == null) 
				uc = new UdpClient(new IPEndPoint(ge.addr, ge.UDPPort));
			if (uc == null)
				return null;
			BinaryFormatter bf = new BinaryFormatter();
			MemoryStream ms = new MemoryStream();
			DatagramWrapper[] dw = new DatagramWrapper[1];
			dw[0] = new DatagramWrapper(o, ge.gridNb, destinations[dest].gridNb);
			bf.Serialize(ms, dw);
			return ms.GetBuffer(); 
		}

		/// <summary>
		/// Send an object to another process in SMTP
		/// </summary>
		/// <param name="o">Object to send</param>
		/// <param name="dest">Destination number</param>
		/// <returns>-1 in case of error</returns>
		private int SMTPSend(object o, int dest) 
		{
			if ((SMTPServer == null) || (destinations[dest].mailAddress == null))
				return -1;
			MailMessage mm = new MailMessage();
			mm.BodyFormat = MailFormat.Text;
			mm.BodyEncoding = Encoding.ASCII;
			mm.To = destinations[dest].mailAddress;
			mm.From = ge.mailAddress;
			mm.Headers.Add("Reply-To", ge.mailAddress);
			mm.Priority = MailPriority.Normal;
			mm.Subject = "PageBoxGrid";
			DatagramWrapper[] dw = new DatagramWrapper[1];
			dw[0] = new DatagramWrapper(o, ge.gridNb, destinations[dest].gridNb);
			BinaryFormatter bf = new BinaryFormatter();
			MemoryStream ms = new MemoryStream();
			bf.Serialize(ms, dw);
			ASCIIEncoding ae = new ASCIIEncoding();
			mm.Body = ae.GetString(ms.GetBuffer()); 
			SmtpMail.SmtpServer = SMTPServer;
			try 
			{
				SmtpMail.Send(mm);
			}
			catch(Exception ex) 
			{
				return -1;
			}
			return 1;
		}

		/// <summary>
		/// Wait for the next received message object
		/// </summary>
		/// <param name="transport_mode">UDP or SMTP</param>
		/// <returns>Message object</returns>
		public Message Receive(Transport transport_mode) 
		{
			switch(transport_mode) 
			{
				case Transport.UDP:
					return UDPReceive(true);
				case Transport.SMTP:
					return SMTPReceive(true);
			}
			return null;
		}

		/// <summary>
		/// Get the next received object
		/// Version 0.0.5d
		/// </summary>
		/// <param name="transport_mode">UDP or SMTP</param>
		/// <returns>Message object</returns>
		public Message ReceiveNonBlock(Transport transport_mode) 
		{
			switch(transport_mode) 
			{
				case Transport.UDP:
					return UDPReceive(false);
				case Transport.SMTP:
					return SMTPReceive(false);
			}
			return null;
		}

		/// <summary>
		/// Wait for the next received message object sent using UDP.
		/// Version 0.0.5d: block parameter.
		/// </summary>
		/// <param name="block">True if the receive must wait</param>
		/// <returns>Message object</returns>
		private Message UDPReceive(bool block) 
		{
			Message m = UDPGetMessage();
			if (!block || (m != null))
				return m;
			// If we are here there was nothing queued.
			lock(this) 
			{
				Monitor.Wait(this, timeout);	
			}
			return UDPGetMessage();
		}

		/// <summary>
		/// Get the next received message object.
		/// </summary>
		/// <returns>Message object</returns>
		private Message UDPGetMessage() 
		{
			Message m = null;
			lock(this) 
			{
				if (clientQueue.Count > 0) 
				{
					receivePending = false;
					m = (Message)clientQueue.Dequeue();
				}
				else
					receivePending = true;
			}
			if ((UDPThread != null) && (m != null)) 
				UDPThread.UnsetReceivePending(ge.gridNb);
			if ((multicastThread != null) && (m != null)) 
				multicastThread.UnsetReceivePending(ge.gridNb);
			if (m != null)
				return m;
			if (UDPThread != null) 
			{
				UDPMessage um = UDPThread.Receive(ge.gridNb);
				if (um != null) 
				{
					m = new Message();
					m.recv = um.o;
					m.transport_mode = Transport.UDP;
					m.from = GetOriginFromUDP(um.addr, um.port, um.gridNb);
					return m;
				}
			}
			if (multicastThread != null) 
			{
				UDPMessage um = multicastThread.Receive(ge.gridNb);
				if (um != null) 
				{
					m = new Message();
					m.recv = um.o;
					m.transport_mode = Transport.UDP;
					m.from = GetOriginFromUDP(um.addr, um.port, um.gridNb);
					return m;
				}
			}
			return null;
		}

		/// <summary>
		/// Retrieves the destination nb from the UDP address and port and from the Grid number.
		/// Called by UDPGetMessage (Grid.cs) and by UDPQueue.Enqueue (UDPQueue.cs).
		/// Version 0.0.5d.
		/// </summary>
		/// <param name="addrFrom">UDP address of the sender</param>
		/// <param name="portFrom">UDP port of the sender</param>
		/// <param name="gridNbFrom">Grid number of the sender</param>
		/// <returns>Destination number of the sender</returns>
		public int GetOriginFromUDP(IPAddress addrFrom, int portFrom, int gridNbFrom) 
		{
			for (int i = 0; i < destinations.Length; ++i) 
			{
				if (destinations[i].addr.Equals(addrFrom) &&
					(destinations[i].UDPPort == portFrom) &&
					(destinations[i].gridNb == gridNbFrom))
					return i;
			}
			return -1;
		}

		/// <summary>
		/// Wait for the next received message object sent using SMTP.
		/// Version 0.0.5d: block parameter.
		/// </summary>
		/// <param name="block">True if the receive must wait</param>
		/// <returns>Message object</returns>
		private Message SMTPReceive(bool block) 
		{
			Message m = SMTPGetMessage();
			if (!block || (m != null))
				return m;
			// If we are here there was nothing queued.
			lock(this) 
			{
				Monitor.Wait(this, timeout);	
			}
			return SMTPGetMessage();
		}

		/// <summary>
		/// Get the next received message object
		/// </summary>
		/// <returns>Message object</returns>
		private Message SMTPGetMessage() 
		{
			Message m = null;
			lock(this) 
			{
				if (clientQueue.Count > 0) 
				{
					receivePending = false;
					m = (Message)clientQueue.Dequeue();
				}
				else
					receivePending = true;
			}
			if (POPThread != null) 
			{
				if (m != null) 
				{
					POPThread.UnsetReceivePending(ge.gridNb);
					return m;
				}
				SMTPMessage sm = POPThread.Receive(ge.gridNb);
				if (sm != null) 
				{
					m = new Message();
					m.recv = sm.o;
					m.transport_mode = Transport.SMTP;
					m.from = GetOriginFromMail(sm.mailAddress, sm.gridNb);
					return m;
				}
			}
			return null;
		}

		/// <summary>
		/// Retrieves the destination nb from the mail address and Grid number.
		/// Called by SMTPGetMessage (Grid.cs) and GridQueue.Enqueue (POPListener.cs).
		/// Version 0.0.5d.
		/// </summary>
		/// <param name="mailAddress">Mail address of the sender</param>
		/// <param name="gridNb">Grid number of the sender</param>
		/// <returns>Destination number of the sender</returns>
		public int GetOriginFromMail(string mailAddress, int gridNb) 
		{
			for (int i = 0; i < destinations.Length; ++i) 
			{
				if (destinations[i].mailAddress.Equals(mailAddress) &&
					(destinations[i].gridNb == gridNb))
					return i;
			}
			return -1;
		}

		/// <summary>
		/// Send a set of objects to other processes in the group.
		/// All objects must be consumed. MIMD form.
		/// </summary>
		/// <param name="os">Array of objects</param>
		/// <param name="transport_mode">UDP or SMTP</param>
		/// <returns>Array of error status</returns>
		public int[] Scatter(object[] os, Transport transport_mode) 
		{
			switch(transport_mode) 
			{
				case Transport.UDP:
					return UDPScatter(os);
				case Transport.SMTP:
					return SMTPScatter(os);
			}
			return null;
		}

		/// <summary>
		/// Broadcast messages using UDP.
		/// </summary>
		/// <param name="os">Array of objects</param>
		/// <returns>Array of error status: -1 in case of error or number of
		/// messages sent to that destination</returns>
		private int[] UDPScatter(object[] os) 
		{
			if ((ge.addr == null) || (ge.UDPPort == 0))
				return null;
			if (uc == null) 
				uc = new UdpClient(new IPEndPoint(ge.addr, ge.UDPPort));
			if (uc == null)
				return null;
			int[] errors = new int[destinations.Length];
			IDictionaryEnumerator ide = UDPDests.GetEnumerator();
			for (int i = 0; i < os.Length; ++ i) 
			{
				if (!ide.MoveNext()) 
					ide.Reset();
				IPEndPoint iep = (IPEndPoint)ide.Key;
				ArrayList al = (ArrayList)ide.Value;
				if ((iep.Address == ge.addr) && (iep.Port == ge.UDPPort)) 
				{
					int j = 0;
					for (; j < al.Count; ++j) 
					{
						int gridNb = (int)al[j];
						if (gridNb == ge.gridNb)
							continue;
						for (int k = 0; k < destinations.Length; ++k) 
						{
							if ((destinations[k].addr == iep.Address) &&
								(destinations[k].UDPPort == iep.Port) &&
								(destinations[k].gridNb == gridNb)) 
							{
								MemorySend(os[i], GridEntry.getGrid(destinations[k]),
									Transport.UDP);
								++ i;
								break;
							}
						}
						if (i == os.Length)
							break;
					}
					for (int k = 0; k < j; ++k)
						++ errors[(int)al[k]];
				} 
				else 
				{
					DatagramWrapper[] dw = new DatagramWrapper[al.Count];
					int j = 0;
					for (; j < dw.Length; ++j) 
					{
						dw[j] = new DatagramWrapper(os[i], ge.gridNb, (int)al[j]);
						++ i;
						if (i == os.Length)
							break;
					}
					BinaryFormatter bf = new BinaryFormatter();
					MemoryStream ms = new MemoryStream();
					bf.Serialize(ms, dw);
					byte[] buf = ms.GetBuffer();
					if (uc.Send(buf, buf.Length, (IPEndPoint)ide.Key) < 0) 
					{
						for (int k = 0; k < j; ++k)
							errors[(int)al[k]] = -1;
					} 
					else 
					{
						for (int k = 0; k < j; ++k)
							++ errors[(int)al[k]];
					}
				}
			}
			return errors;
		}

		/// <summary>
		/// Broadcast messages using SMTP.
		/// </summary>
		/// <param name="os">Array of objects</param>
		/// <returns>Array of error status: -1 in case of error or number of
		/// messages sent to that destination</returns>
		private int[] SMTPScatter(object[] os) 
		{
			if (SMTPServer == null)
				return null;
			int[] errors = new int[destinations.Length];
			IDictionaryEnumerator ide = mailDests.GetEnumerator();
			for (int i = 0; i < os.Length; ++ i) 
			{
				if (!ide.MoveNext()) 
					ide.Reset();
				ArrayList al = (ArrayList)ide.Value;
				string mailAddress = (string)ide.Key;
				if (mailAddress.Equals(ge.mailAddress)) 
				{
					int j = 0;
					for (; j < al.Count; ++j) 
					{
						int gridNb = (int)al[j];
						if (gridNb == ge.gridNb)
							continue;
						for (int k = 0; k < destinations.Length; ++k) 
						{
							if ((destinations[k].mailAddress.Equals(mailAddress)) &&
								(destinations[k].gridNb == gridNb)) 
							{
								MemorySend(os[i], GridEntry.getGrid(destinations[k]), 
									Transport.SMTP);
								++ i;
								break;
							}
						}
						if (i == os.Length)
							break;
					}
					for (int k = 0; k < j; ++k)
						++ errors[(int)al[k]];
				} 
				else 
				{
					DatagramWrapper[] dw = new DatagramWrapper[al.Count];
					int j = 0;
					for (; j < dw.Length; ++j) 
					{
						dw[j] = new DatagramWrapper(os[i], ge.gridNb, (int)al[j]);
						++ i;
						if (i == os.Length)
							break;
					}
					BinaryFormatter bf = new BinaryFormatter();
					MemoryStream ms = new MemoryStream();
					bf.Serialize(ms, dw);
					MailMessage mm = new MailMessage();
					mm.BodyFormat = MailFormat.Text;
					mm.BodyEncoding = Encoding.ASCII;
					mm.To = (string)ide.Key;
					mm.From = ge.mailAddress;
					mm.Headers.Add("Reply-To", ge.mailAddress);
					mm.Priority = MailPriority.Normal;
					mm.Subject = "PageBoxGrid";
					ASCIIEncoding ae = new ASCIIEncoding();
					mm.Body = ae.GetString(ms.GetBuffer()); 
					SmtpMail.SmtpServer = SMTPServer;
					try 
					{
						SmtpMail.Send(mm);
					}
					catch(Exception ex) 
					{
						// Returns an error only when the SMTP server is not working
						return null;
					}
					for (int k = 0; k < j; ++k)
						++ errors[(int)al[k]];
				}
			}
			return errors;
		}

		/// <summary>
		/// Send an object to all other processes in the group. SPMD form.
		/// </summary>
		/// <param name="os">Objects</param>
		/// <param name="transport_mode">UDP or SMTP</param>
		/// <returns>Error status</returns>
		public int Scatter(object o, Transport transport_mode) 
		{
			switch(transport_mode) 
			{
				case Transport.UDP:
					return UDPScatter(o);
				case Transport.SMTP:
					return SMTPScatter(o);
			}
			return -1;
		}

		/// <summary>
		/// Send an object to all other processes in the group in UDP. SPMD form.
		/// We use multicast if possible.
		/// </summary>
		/// <param name="o">Object to send</param>
		/// <returns>Error status. When multicast is not possible, -1 if at least one error.</returns>
		private int UDPScatter(object o) 
		{
			byte[] buf;
			if (multicastAddress != null) 
			{
				if (ucm == null) 
				{
					ucm = new UdpClient(multicastPort);
					ucm.JoinMulticastGroup(multicastAddress, ttl);
				}
				if (ucm == null)
					return -1;
				if (multicastEP == null)
					multicastEP = new IPEndPoint(multicastAddress, multicastPort);
				BinaryFormatter bf = new BinaryFormatter();
				MemoryStream ms = new MemoryStream();
				MulticastWrapper dw = new MulticastWrapper(o, ge.gridNb, ge.addr, ge.UDPPort);
				bf.Serialize(ms, dw);
				buf = ms.GetBuffer(); 
				if (ucm.Send(buf, buf.Length, multicastEP) < 0)
					return -1;
				return 1;
			}
			int rc = 1;
			// -2 for all destinations
			buf = UDPPrepare(o, -2);
			IDictionaryEnumerator ide = UDPDests.GetEnumerator();
			while(ide.MoveNext()) 
			{
				IPEndPoint iep = (IPEndPoint)ide.Key;
				if ((iep.Address == ge.addr) && (iep.Port == ge.UDPPort)) 
				{
					ArrayList al = (ArrayList)ide.Value;
					for (int i = 0; i < al.Count; ++i) 
					{
						int gridNb = (int)al[i];
						if (gridNb == ge.gridNb)
							continue;
						for (int j = 0; j < destinations.Length; ++j) 
						{
							if ((destinations[j].addr == iep.Address) &&
								(destinations[j].UDPPort == iep.Port) &&
								(destinations[j].gridNb == gridNb)) 
							{
								MemorySend(o, GridEntry.getGrid(destinations[j]), 
									Transport.UDP);
								break;
							}
						}
					}
				}
				else
					if (uc.Send(buf, buf.Length, iep) < 0)
						return -1;
			}
			return rc;
		}

		/// <summary>
		/// Send an object to all other processes in the group in SMTP. SPMD form.
		/// </summary>
		/// <param name="o">Object to send</param>
		/// <returns>Error status. When multicast is not possible, -1 if at least one error.</returns>
		private int SMTPScatter(object o) 
		{
			if (SMTPServer == null)
				return -1;
			MailMessage mm = new MailMessage();
			mm.BodyFormat = MailFormat.Text;
			mm.BodyEncoding = Encoding.ASCII;
			IDictionaryEnumerator ide = mailDests.GetEnumerator();
			bool first = true;
			while (ide.MoveNext()) 
			{
				string mailAddress = (string)ide.Key;
				if (mailAddress.Equals(ge.mailAddress)) 
				{
					ArrayList al = (ArrayList)ide.Value;
					for (int i = 0; i < al.Count; ++i) 
					{
						int gridNb = (int)al[i];
						if (gridNb == ge.gridNb)
							continue;
						for (int j = 0; j < destinations.Length; ++j) 
						{
							if ((destinations[j].mailAddress.Equals(mailAddress)) &&
								(destinations[j].gridNb == gridNb)) 
							{
								MemorySend(o, GridEntry.getGrid(destinations[j]), 
									Transport.SMTP);
								break;
							}
						}
					}
				}
				if (first) 
				{
					mm.To = mailAddress;
					first = false;
				} else
					mm.To += (";" + mailAddress);
			}
			mm.From = ge.mailAddress;
			mm.Headers.Add("Reply-To", ge.mailAddress);
			mm.Priority = MailPriority.Normal;
			mm.Subject = "PageBoxGrid";
			DatagramWrapper[] dw = new DatagramWrapper[1];
			// -2 stands here for all
			dw[0] = new DatagramWrapper(o, ge.gridNb, -2);
			BinaryFormatter bf = new BinaryFormatter();
			MemoryStream ms = new MemoryStream();
			bf.Serialize(ms, dw);
			ASCIIEncoding ae = new ASCIIEncoding();
			mm.Body = ae.GetString(ms.GetBuffer()); 
			SmtpMail.SmtpServer = SMTPServer;
			try 
			{
				SmtpMail.Send(mm);
			}
			catch(Exception ex) 
			{
				return -1;
			}
			return 1;
		}

		/// <summary>
		/// Waits for a set of objects coming from other processes
		/// </summary>
		/// <param name="errors">null (all) or an error array returned by a former broadcast</param>
		/// <param name="transport_mode">UDP or SMTP</param>
		/// <returns>A Message array</returns>
		public Message[] Gather(ref int[] errors, Transport transport_mode) 
		{
			switch(transport_mode) 
			{
				case Transport.UDP:
					return UDPGather(ref errors);
				case Transport.SMTP:
					return SMTPGather(ref errors);
			}
			return null;
		}

		/// <summary>
		/// Waits for a set of objects coming from other processes and sent using UDP
		/// </summary>
		/// <param name="errors">null (all) or an error array returned by a former broadcast</param>
		/// <returns>A Message array</returns>
		private Message[] UDPGather(ref int[] errors) 
		{
			ArrayList al = new ArrayList();
			if (UDPGather2(ref al, ref errors))
				return (Message [])al.ToArray(typeof(Message));
			lock(this) 
			{
				Monitor.Wait(this, timeout);	
			}
			// Best effort:
			// Returns anyway what we have found
			UDPGather2(ref al, ref errors);
			return (Message [])al.ToArray(typeof(Message));
		}

		/// <summary>
		/// Get the next messages in queue up to the number expected in the array of errors.
		/// </summary>
		/// <param name="al">Array of messages</param>
		/// <param name="founds">Array of errors. UDPGather2 updates the array.</param>
		/// <returns>true if the contract was fulfilled</returns>
		private bool UDPGather2(ref ArrayList al, ref int[]founds) 
		{
			Message m = null;
			int i;
			while((m = UDPGetMessage()) != null) 
			{
				al.Add(m);
				if ((m.from != -1) && (founds[m.from] > 0)) 
					-- founds[m.from];
			}
			for (i = 0; i < founds.Length; ++i) 
			{
				if (founds[i] > 0)
					return false;
			}
			return true;
		}

		/// <summary>
		/// Waits for a set of objects coming from other processes and sent using SMTP
		/// </summary>
		/// <param name="errors">Error array returned by a former scatter.
		/// SMTPGather updates the array.</param>
		/// <returns>A Message array</returns>
		private Message[] SMTPGather(ref int[] errors) 
		{
			ArrayList al = new ArrayList();
			if (SMTPGather2(ref al, ref errors))
				return (Message [])al.ToArray(typeof(Message));
			lock(this) 
			{
				Monitor.Wait(this, timeout);	
			}
			// Best effort:
			// Returns anyway what we have found
			SMTPGather2(ref al, ref errors);
			return (Message [])al.ToArray(typeof(Message));
		}

		/// <summary>
		/// Get the next messages in queue up to the number expected in the array of errors.
		/// </summary>
		/// <param name="al">Array of messages</param>
		/// <param name="founds">Array of errors. SMTPGather2 updates the array.</param>
		/// <returns>true if the contract was fulfilled</returns>
		private bool SMTPGather2(ref ArrayList al, ref int[]founds) 
		{
			Message m = null;
			int i;
			while((m = SMTPGetMessage()) != null) 
			{
				al.Add(m);
				if ((m.from != -1) && (founds[m.from] > 0)) 
					-- founds[m.from];
			}
			for (i = 0; i < founds.Length; ++i) 
			{
				if (founds[i] > 0)
					return false;
			}
			return true;
		}

		/// <summary>
		/// Convenience method to build an error array to use in Gather.
		/// </summary>
		/// <param name="val">-1/0 nothing expected or nb of messages expected</param>
		/// <returns>Array of errors</returns>
		public int[] GetDestinations(int val) 
		{
			int[] errors = new int[destinations.Length];
			for (int i = 0; i < errors.Length; ++i)
				errors[i] = val;
			return errors;
		}

		/// <summary>
		/// Retrieve extra parameter of destinations.
		/// Version 0.0.5d
		/// </summary>
		/// <returns>An array of destination extra parameters</returns>
		public string[] GetParms() 
		{
			string[] parms = new string[destinations.Length];
			for (int i = 0; i < parms.Length; ++i)
				parms[i] = destinations[i].parm;
			return parms;
		}

		/// <summary>
		/// Retrieve the extra parameter of a destination
		/// Version 0.0.5d
		/// </summary>
		/// <param name="dest">Destination number</param>
		/// <returns>Extra parameter</returns>
		public string GetParm(int dest) 
		{
			return destinations[dest].parm;
		}

		/// <summary>
		/// Retrieve the destinations with an extra parameter
		/// Version 0.0.5d
		/// </summary>
		/// <param name="parm">Extra parameter</param>
		/// <returns>Array of destination numbers</returns>
		public int[] GetDestFromParm(string parm) 
		{
			ArrayList al = new ArrayList();
			for (int i = 0; i < destinations.Length; ++i) 
			{
				if (parm.Equals(destinations[i].parm))
					al.Add(i);
			}
			return (int[])al.ToArray(typeof(int));
		}

		/// <summary>
		/// Subscribe a message handler implementing the GridCallback interface.
		/// </summary>
		/// <param name="gc">Message handler</param>
		/// Version 0.0.5d
		public void Subscribe(GridCallback gc) 
		{
			this.gc = gc;
		}

		/// <summary>
		/// To use when a peer fails to answer.
		/// 1. Notifies peers using the fastest available mean: multicast first and then UDP and SMTP
		/// 2. Updates ActiveNaming
		/// </summary>
		/// <param name="dest">Destination to remove</param>
		/// <returns>true: success</returns>
		public bool RemoveDestination(int dest) 
		{
			if ((dest < 0) || (dest >= destinations.Length))
				return false;
			GridEntry ger = (GridEntry)destinations[dest];
			lock(this) 
			{
				ArrayList al = null;
				if (ger.mailAddress != null) 
				{
					al = (ArrayList)mailDests[ger.mailAddress];
					al.Remove(dest);
					if (al.Count == 0)
						mailDests.Remove(ger.mailAddress);
				}
				if ((ger.UDPAddress != null) && (ger.UDPPort != 0)) 
				{
					IPEndPoint iep = new IPEndPoint(ger.addr, ger.UDPPort);
					al = (ArrayList)UDPDests[iep];
					al.Remove(dest);
					if (al.Count == 0)
						UDPDests.Remove(iep);
				}
				al = new ArrayList(destinations);
				al.RemoveAt(dest);
				destinations = (GridEntry[])al.ToArray(typeof(GridEntry));
			}
			GridCommand gc = new GridCommand(ger, Command.Remove);
			if (Scatter(gc, Transport.UDP) == -1)
				Scatter(gc, Transport.SMTP);
			ActiveProxy ap = new ActiveProxy(gc.activeNamingUrl);
			try 
			{
				ap.Discover();
				ArrayList al = deserializeGEs(ap.GetKeyRange("PageBoxGrid.dll", communicator));
				al.Remove(ger);
				ap.SetKeyRange("PageBoxGrid.dll", communicator, serializeGEs(al));
			}
			catch(Exception e) 
			{
				log.write("Grid.RemoveDestination(" + gc.activeNamingUrl + ", " + gc.UDPAddress +
					", " + gc.UDPPort + ", " + gc.mailAddress + ", " + gc.gridNb + ") returned " +
					e.Message);
				return false;
			}
			return true;
		}

		/// <summary>
		/// Deserialize GridEntry ArrayList from ActiveNaming KeyRange
		/// </summary>
		/// <param name="keyRange">Serialized key range</param>
		/// <returns>Deserialized ArrayList</returns>
		ArrayList deserializeGEs(string keyRange) {
			BinaryFormatter bf = new BinaryFormatter();
			ASCIIEncoding ae = new ASCIIEncoding();
			MemoryStream ms = new MemoryStream(ae.GetBytes(keyRange));
			return (ArrayList)bf.Deserialize(ms);
		}

		/// <summary>
		/// Serialize GridEntry ArrayList to string
		/// </summary>
		/// <param name="al">GridEntry ArrayList</param>
		/// <returns>Serialized key range</returns>
		string serializeGEs(ArrayList al) 
		{
			ASCIIEncoding ae = new ASCIIEncoding();
			BinaryFormatter bf = new BinaryFormatter();
			MemoryStream ms = new MemoryStream();
			bf.Serialize(ms, al);
			return ae.GetString(ms.GetBuffer());
		}
	}
	/// <summary>
	/// A grid entry uniquely identifies a destination.
	/// At the grid creation time, the user must set for SMTP/POP
	/// - the mail address
	/// - the corresponding POP server IP address, user and password
	/// Restriction:
	/// Thank to the grid number any number of threads in the same OS process
	/// can share the same mail address. However you need at least one mail address
	/// per process.
	/// (mail address, grid number) must be unique, which implies that two OS processes
	/// must not use the same mail address.
	/// </summary>
	[Serializable()]
	public class GridEntry 
	{
		/// <summary>
		/// Destination UDP address
		/// </summary>
		public string UDPAddress;
		/// <summary>
		/// IPAddress generated from UDPAddress
		/// </summary>
		public IPAddress addr;
		/// <summary>
		/// Destination UDP port
		/// </summary>
		public int UDPPort;
		/// <summary>
		/// Destination mail address
		/// </summary>
		public string mailAddress;
		/// <summary>
		/// Extra parameter of the Grid Entry
		/// </summary>
		public string parm;
		/// <summary>
		/// Destination grid number
		/// </summary>
		public int gridNb;
		/// <summary>
		/// URL of the controlling ActiveNaming
		/// </summary>
		public string activeNamingUrl;
		/// <summary>
		/// Key: mail address
		/// Value: current grid number for this address
		/// </summary>
		private static Hashtable grids = new Hashtable();
		/// <summary>
		/// Key: address
		/// Value: ArrayList of Grid objects
		/// </summary>
		private static Hashtable processGrids = new Hashtable();
		/// <summary>
		/// GridEntry factory. Manages the destination grid number.
		/// Also manages processGrids which is used for thread to thread communication.
		/// </summary>
		/// <param name="UDPAddress">UDP address</param>
		/// <param name="UDPPort">UDP port</param>
		/// <param name="mailAddress">Mail address of the destination</param>
		/// <param name="url">URL of the controlling ActiveNaming</param>
		/// <param name="parm">Extra GridEntry parameter</param>
		/// <param name="grid">Creating Grid object</param>
		/// <returns>Created Grid Entry</returns>
		public static GridEntry getGridEntry(string UDPAddress, int UDPPort, 
			string mailAddress, string url, string parm, Grid grid) 
		{
			lock(typeof(GridEntry)) 
			{
				GridEntry ge;
				string key = GenKey(UDPAddress, UDPPort, mailAddress);
				if (grids.ContainsKey(key)) 
				{
					int gridNb = (int)grids[key] + 1;
					grids[key] = gridNb;
					ge = new GridEntry(UDPAddress, UDPPort, mailAddress, gridNb, url, parm);
					ArrayList al = (ArrayList)processGrids[key];
					al.Insert(gridNb, grid);
				} 
				else 
				{
					grids.Add(key, (int)0);
					ge = new GridEntry(UDPAddress, UDPPort, mailAddress, 0, url, parm);
					ArrayList al = new ArrayList();
					al.Insert(0, grid);
					processGrids.Add(key, al);
				}
				return ge;
			}
		}

		/// <summary>
		/// Generates a key made of the UDP address, port and mail address.
		/// </summary>
		/// <param name="UDPAddress">UDP address</param>
		/// <param name="UDPPort">UDP port</param>
		/// <param name="mailAddress">Mail address of the destination</param>
		/// <returns>Key</returns>
		private static string GenKey(string UDPAddress, int UDPPort, string mailAddress) 
		{
			string key = null;
			if (UDPAddress != null) 
			{
				key = UDPAddress;
				if (UDPPort != 0)
					key += (":" + UDPPort);
			}
			if (mailAddress != null) 
			{
				if (key == null)
					key = mailAddress;
				else
					key += ("&" + mailAddress);
			}
			return key;
		}

		/// <summary>
		/// When the destination is in the same OS process, returns its Grid object.
		/// Used for memory communication.
		/// </summary>
		/// <param name="ge">destination</param>
		/// <returns>Grid object or null</returns>
		public static Grid getGrid(GridEntry ge) 
		{
			lock(typeof(GridEntry)) 
			{
				string key = GenKey(ge.UDPAddress, ge.UDPPort, ge.mailAddress);
				if (!processGrids.Contains(key))
					return null;
				ArrayList al = (ArrayList)processGrids[key];
				return (Grid)al[ge.gridNb];
			}
		}

		/// <summary>
		/// Constructor. Sets the mail address and grid number.
		/// A destination is made of (mailaddress, grid number).
		/// </summary>
		/// <param name="UDPAddress">UDP address</param>
		/// <param name="UDPPort">UDP port</param>
		/// <param name="mailAddress">Destination mail address</param>
		/// <param name="gridNb">Destination grid number</param>
		/// <param name="activeNamingUrl">URL of the controlling ActiveNaming</param>
		/// <param name="parm">Extra Grid parameter</param>
		public GridEntry(string UDPAddress, int UDPPort, string mailAddress, int gridNb,
			string activeNamingUrl, string parm) 
		{
			this.UDPAddress = UDPAddress;
			IPHostEntry ihe = Dns.GetHostByName(UDPAddress);
			addr = ihe.AddressList[0];
			this.UDPPort = UDPPort;
			this.mailAddress = mailAddress;
			this.gridNb = gridNb;
			this.activeNamingUrl = activeNamingUrl;
			this.parm = parm;
		}
	}

	/// <summary>
	/// Represents what is returned by the API Receive and Gather.
	/// </summary>
	public class Message 
	{
		/// <summary>
		/// Received object
		/// </summary>
		public object recv;
		/// <summary>
		/// Destination that sent the object
		/// </summary>
		public int from;
		/// <summary>
		/// Transport mode used to send the object
		/// </summary>
		public Transport transport_mode;
	}

	/// <summary>
	/// System commands supported by PageBox Grid
	/// </summary>
	public enum Command 
	{
		Remove,
		Add
	}

	/// <summary>
	/// Object used to pass dynamic add and remove commands.
	/// </summary>
	public class GridCommand 
	{
		/// <summary>
		/// Command to run
		/// </summary>
		public Command cmd;
		/// <summary>
		/// Destination UDP address
		/// </summary>
		public string UDPAddress;
		/// <summary>
		/// Destination IP address
		/// </summary>
		public IPAddress addr;
		/// <summary>
		/// Destination UDP port
		/// </summary>
		public int UDPPort;
		/// <summary>
		/// Destination mail address
		/// </summary>
		public string mailAddress;
		/// <summary>
		/// Destination grid number
		/// </summary>
		public int gridNb;
		/// <summary>
		/// URL of the controlling ActiveNaming
		/// </summary>
		public string activeNamingUrl;
		/// <summary>
		/// Extra Grid parameter
		/// </summary>
		public string parm;

		/// <summary>
		/// Constructor. Set member variables
		/// </summary>
		/// <param name="ge">GridEntry target</param>
		/// <param name="cmd">Command to run</param>
		public GridCommand(GridEntry ge, Command cmd) 
		{
			this.cmd = cmd;
			UDPAddress = ge.UDPAddress;
			addr = ge.addr;
			UDPPort = ge.UDPPort;
			mailAddress = ge.mailAddress;
			gridNb = ge.gridNb;
			activeNamingUrl = ge.activeNamingUrl;
			parm = ge.parm;
		}

		/// <summary>
		/// Checks if a GridEntry has the same content.
		/// </summary>
		/// <param name="ge">GridEntry to compare with</param>
		/// <returns>true if same content</returns>
		public bool Matches(GridEntry ge) 
		{
			return (UDPAddress.Equals(ge.UDPAddress) && addr.Equals(ge.addr) &&
				(UDPPort == ge.UDPPort) && mailAddress.Equals(ge.mailAddress) &&
				(gridNb == ge.gridNb) && activeNamingUrl.Equals(ge.activeNamingUrl));
		}

		/// <summary>
		/// Apply the command to the target process
		/// </summary>
		/// <param name="grid">Target Grid object</param>
		public void Update(Grid grid) 
		{
			lock(grid) 
			{
				ArrayList al = null;
				switch(cmd)
				{
					case Command.Add:
						if (mailAddress != null) 
						{
							al = (ArrayList)grid.mailDests[mailAddress];
							al.Add(gridNb);
						}
						if ((UDPAddress != null) && (UDPPort != 0)) 
						{
							al = (ArrayList)grid.UDPDests[new IPEndPoint(addr, UDPPort)];
							al.Add(gridNb);
						}
						al = new ArrayList(grid.destinations);
						al.Add(getGridEntry());
						break;
					case Command.Remove: 
						if (mailAddress != null) 
						{
							al = (ArrayList)grid.mailDests[mailAddress];
							al.Remove(gridNb);
						}
						if ((UDPAddress != null) && (UDPPort != 0)) 
						{
							al = (ArrayList)grid.UDPDests[new IPEndPoint(addr, UDPPort)];
							al.Remove(gridNb);
						}
						al = new ArrayList();
						for (int i = 0; i < grid.destinations.Length; ++i) 
						{
							if (Matches(grid.destinations[i]))
								continue;
							al.Add(grid.destinations[i]);
						}
						break;
				}
				if (al != null)
					grid.destinations = (GridEntry[])al.ToArray(typeof(GridEntry));
			}
		}

		/// <summary>
		/// Build a GridEntry from the object content
		/// </summary>
		/// <returns>Built GridEntry</returns>
		private GridEntry getGridEntry() 
		{
			return new GridEntry(UDPAddress, UDPPort, mailAddress, gridNb, activeNamingUrl, parm);
		}
	}
	/// <summary>
	/// Interface provided by the subscriber.
	/// Version 0.0.5d
	/// </summary>
	public interface GridCallback 
	{
		/// <summary>
		/// Notify an event to a subscriber
		/// </summary>
		/// <param name="transport_mode">UDP, SMTP or Memory</param>
		/// <param name="from">Destination number of the issuer</param>
		/// <param name="o">Object to process</param>
		void Notify(Transport transport_mode, int from, object o); 
	}
}
