using System;
using System.Collections;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;

namespace PageBoxGrid
{
	/// <summary>
	/// <p>Multicast Listener.
	/// See <a href="http://pagebox.net/grid.html">Grid</a> for explanations.</p>
	/// <p>Copyright (c) 2002 Alexis Grandemange<br/>
	/// Mail: alexis.grandemange@pagebox.net</p>
	/// <pre>This program is free software; you can redistribute it and/or
	/// modify it under the terms of the GNU Lesser General Public
	/// License as published by the Free Software Foundation; version
	/// 2.1 of the License.
	/// This library is distributed in the hope that it will be useful,
	/// but WITHOUT ANY WARRANTY; without even the implied warranty of
	/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	/// GNU Lesser General Public License for more details.
	/// A copy of the GNU Lesser General Public License lesser.txt should be
	/// included in the distribution.</pre>
	/// </summary>
	public class MulticastListener
	{
		/// <summary>
		/// Key: UDPAddressPort of the UDP server
		/// Value: UDPListener instance
		/// </summary>
		private static Hashtable servers = new Hashtable();
		/// <summary>
		/// Key: Sequence number of the Grid client
		/// Value: UDPQueue
		/// </summary>
		private Hashtable clients = new Hashtable();
		/// <summary>
		/// Log object used to write errors on a file
		/// </summary>
		private Log log;
		/// <summary>
		/// When true, stops the thread
		/// </summary>
		private bool toStop = false;
		/// <summary>
		/// UDP address
		/// </summary>
		private string UDPAddress;
		/// <summary>
		/// UDP Port
		/// </summary>
		private int UDPPort;
		/// <summary>
		/// Time to live
		/// </summary>
		private int ttl;
		/// <summary>
		/// Listener thread
		/// </summary>
		private Thread listenerThread;

		/// <summary>
		/// Static method that retrieves or creates a UDP Listener.
		/// </summary>
		/// <param name="UDPAddress">multicast address</param>
		/// <param name="UDPPort">multicast port</param>
		/// <param name="ttl">Time to live</param>
		/// <param name="log">Log object used to write errors on a file</param>
		/// <param name="grid">Grid client</param>
		/// <returns>Allocated Multicast Listener</returns>
		public static MulticastListener GetListener(string UDPAddress, int UDPPort, int ttl,
			Log log, Grid grid) 
		{
			lock(typeof(MulticastListener)) 
			{
				MulticastListener ml;
				UDPAddressPort uap = new UDPAddressPort(UDPAddress, UDPPort);
				if (servers.Contains(uap)) 
				{
					ml = (MulticastListener)servers[uap];
					ml.addClient(grid);
				} 
				else 
				{
					ml = new MulticastListener(UDPAddress, UDPPort, ttl, log, grid);
					servers.Add(uap, ml);
					ml.listenerThread = new Thread(new ThreadStart(ml.ThreadRun));
					ml.listenerThread.Start();
				}
				return ml;
			}
		}

		/// <summary>
		/// Constructor. Set member variables.
		/// </summary>
		/// <param name="UDPAddress">UDP address</param>
		/// <param name="UDPPort">UDP port</param>
		/// <param name="log">Log object used to write errors on a file</param>
		/// <param name="grid">Grid client</param>
		public MulticastListener(string UDPAddress, int UDPPort, int ttl, Log log, Grid grid)
		{
			this.UDPAddress = UDPAddress;
			this.UDPPort = UDPPort;
			this.log = log;
			this.ttl = ttl;
			addClient(grid);
		}

		/// <summary>
		/// Add a client to clients
		/// </summary>
		/// <param name="grid">Client object</param>
		private void addClient(Grid grid) 
		{
			lock(this) 
			{
				clients.Add(grid.ge.gridNb, new UDPQueue(grid));
			}
		}

		/// <summary>
		/// Main loop of the MulticastListener thread.
		/// </summary>
		public void ThreadRun() 
		{
			UdpClient uc = new UdpClient(UDPPort);
			uc.JoinMulticastGroup(IPAddress.Parse(UDPAddress), ttl);
			IPEndPoint remoteEP = null;
			while(!toStop) 
			{
				byte[] buf = uc.Receive(ref remoteEP);
				QueueMessage(buf);
			}
		}

		/// <summary>
		/// Stops the thread
		/// </summary>
		public void stop() 
		{
			lock(this) 
			{
				toStop = true;
			}
		}

		/// <summary>
		/// Unserialize and queue a message.
		/// </summary>
		/// <param name="remoteAddr">IP address of the origin</param>
		/// <param name="remotePort">UDP port of the origin</param>
		/// <param name="buf">Message received</param>
		private void QueueMessage(byte[] buf) 
		{
			BinaryFormatter bf = new BinaryFormatter();
			MemoryStream ms = new MemoryStream(buf);
			MulticastWrapper mw = (MulticastWrapper)bf.Deserialize(ms);
			if (mw.o is GridCommand) 
			{
				GridCommand gc = (GridCommand)mw.o;
				IDictionaryEnumerator ide = clients.GetEnumerator();
				while(ide.MoveNext()) 
				{
					UDPQueue uq = (UDPQueue)ide.Value;
					gc.Update(uq.client);
				}
				return;
			}
			lock(this) 
			{
				IDictionaryEnumerator ide = clients.GetEnumerator();
				while(ide.MoveNext()) 
				{
					UDPQueue uq = (UDPQueue)ide.Value;
					uq.Enqueue(mw.o, mw.addrFrom, mw.portFrom, mw.gridNbFrom);
				}
			}
		}

		/// <summary>
		/// Called from Grid.UDPReceive. Calls UDPQueue.Dequeue.
		/// </summary>
		/// <param name="gridNb">Grid number of the destination</param>
		/// <returns>Message wrapped with its origin</returns>
		public UDPMessage Receive(int gridNb) 
		{
			lock(this) 
			{
				UDPQueue gq = (UDPQueue)clients[gridNb];
				return gq.Dequeue();
			}
		}

		/// <summary>
		/// Called from Grid.UDPGetMessage. Unsets UDPQueue's receivePending.
		/// </summary>
		/// <param name="gridNb">Grid number of the destination</param>
		public void UnsetReceivePending(int gridNb) 
		{
			lock(this) 
			{
				UDPQueue gq = (UDPQueue)clients[gridNb];
				gq.UnsetReceivePending();
			}
		}
	}
}
