using System;
using System.Collections;
using System.Threading;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Text;
using System.Runtime.Serialization.Formatters.Binary;

namespace PageBoxGrid
{
	/// <summary>
	/// <p>POP Listener.
	/// See <a href="http://pagebox.net/grid.html">Grid</a> for explanations.</p>
	/// <p>Copyright (c) 2002 Alexis Grandemange<br/>
	/// Mail: alexis.grandemange@pagebox.net</p>
	/// <pre>This program is free software; you can redistribute it and/or
	/// modify it under the terms of the GNU Lesser General Public
	/// License as published by the Free Software Foundation; version
	/// 2.1 of the License.
	/// This library is distributed in the hope that it will be useful,
	/// but WITHOUT ANY WARRANTY; without even the implied warranty of
	/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	/// GNU Lesser General Public License for more details.
	/// A copy of the GNU Lesser General Public License lesser.txt should be
	/// included in the distribution.</pre>
	/// </summary>
	public class POPListener
	{
		/// <summary>
		/// Key: POP3 Server addr + ":" + POP3 User
		/// Value: POPListener instance
		/// </summary>
		private static Hashtable servers = new Hashtable();
		/// <summary>
		/// Key: Sequence number of the Grid client
		/// Value: GridQueue
		/// </summary>
		private Hashtable clients = new Hashtable();
		/// <summary>
		/// Listener thread
		/// </summary>
		private Thread listenerThread;
		/// <summary>
		/// IP address of the POP server
		/// </summary>
		private string POPServer;
		/// <summary>
		/// POP user name
		/// </summary>
		private string POPUser;
		/// <summary>
		/// POP password
		/// </summary>
		private string POPPassword;
		/// <summary>
		/// Log object used to write errors on a file.
		/// </summary>
		private Log log;
		/// <summary>
		/// When true, stops the thread
		/// </summary>
		private bool toStop = false;
		/// <summary>
		/// Stream used to communicate with the POP server
		/// </summary>
		private NetworkStream ns;
		/// <summary>
		/// Each timeout the thread checks for new messages
		/// </summary>
		TimeSpan timeout;
 
		/// <summary>
		/// Static method that retrieves or creates a POP Listener.
		/// </summary>
		/// <param name="POPServer">IP address of the POP server</param>
		/// <param name="POPUser">POP user name. It must match the mail address</param>
		/// <param name="POPPassword">POP password</param>
		/// <param name="log">Log object used to write errors on a file</param>
		/// <param name="timeout">Each timeout minutes the thread checks for new messages</param>
		/// <param name="grid">Grid client</param>
		/// <returns>Allocated POPListener</returns>
		public static POPListener GetListener(string POPServer, string POPUser, string POPPassword, 
			Log log, double timeout, Grid grid) 
		{
			string psk = POPServer + ":" + POPUser;
			lock(typeof(POPListener)) 
			{
				POPListener pl;
				if (servers.Contains(psk)) 
				{ 
					pl = (POPListener)servers[psk];
					pl.addClient(grid);
				}
				else 
				{
					pl = new POPListener(POPServer, POPUser, POPPassword, log, timeout, grid);
					servers.Add(psk, pl);
					pl.listenerThread = new Thread(new ThreadStart(pl.ThreadRun));
					pl.listenerThread.Start();
				}
				return pl;
			}
		}

		/// <summary>
		/// Constructor. Set member variables.
		/// </summary>
		/// <param name="POPServer">IP address of the POP server</param>
		/// <param name="POPUser">POP user name</param>
		/// <param name="POPPassword">POP password</param>
		/// <param name="log">Log object used to write errors on a file</param>
		/// <param name="timeout">Each timeout minutes the thread checks for new messages</param>
		/// <param name="grid">Grid client</param>
		public POPListener(string POPServer, string POPUser, string POPPassword, Log log,
			double timeout, Grid grid)
		{
			this.POPServer = POPServer;
			this.POPUser = POPUser;
			this.POPPassword = POPPassword;
			this.log = log;
			this.timeout = TimeSpan.FromMinutes(timeout);
			addClient(grid);
		}

		/// <summary>
		/// Add a client to clients
		/// </summary>
		/// <param name="grid">Client object</param>
		private void addClient(Grid grid) 
		{
			lock(this) 
			{
				clients.Add(grid.ge.gridNb, new GridQueue(grid));
			}
		}

		/// <summary>
		/// Main loop of the POPListener thread.
		/// Calls Connect, GetMessages and Disconnect.
		/// </summary>
		public void ThreadRun() 
		{
			while(!toStop) 
			{
				if (Connect()) 
				{
					GetMessages();
					Disconnect();
				}
				lock(this) {
					Monitor.Wait(this, timeout);
				}
			}
		}

		/// <summary>
		/// Stops the thread
		/// </summary>
		public void stop() 
		{
			lock(this) 
			{
				toStop = true;
			}
		}

		/// <summary>
		/// We used sample code from Agus Kurniawan and Bill Dean.
		/// It conforms to RFC 1725. 
		/// </summary>
		/// <returns>True in case of success</returns>
		private bool Connect()
		{
			TcpClient server = new TcpClient(POPServer, 110);
			try
			{
				ns = server.GetStream();
				StreamReader sr = new StreamReader(ns);
				sr.ReadLine();
				string data = "USER " + POPUser + "\r\n";
				Byte[] outbytes = System.Text.Encoding.ASCII.GetBytes(data.ToCharArray());
				ns.Write(outbytes, 0, outbytes.Length) ;
				string resp = sr.ReadLine();
				if (!resp.StartsWith("+OK")) 
				{
					log.write("POPListener.Connect " + resp);
					ns.Close();
					return false;
				}
				data = "PASS "+ POPPassword + "\r\n";
				outbytes = System.Text.Encoding.ASCII.GetBytes(data.ToCharArray());
				ns.Write(outbytes, 0, outbytes.Length) ;
				resp = sr.ReadLine();
				if (!resp.StartsWith("+OK")) 
				{
					log.write("POPListener.Connect " + resp);
					ns.Close();
					return false;
				}
			}
			catch(InvalidOperationException ioe)
			{
				log.write("POPListener.Connect unable to connect to " + POPServer);
				return false;
			}
			return true;
		}

		/// <summary>
		/// Separator used in GetMessages and ProcessMessage splits
		/// </summary>
		private char[] blankSep = { ' ' };

		/// <summary>
		/// Get new messages from the POP server.
		/// Calls ProcessMessage.
		/// </summary>
		private void GetMessages() 
		{
			string data = "STAT" + "\r\n";
			Byte[] outbytes = System.Text.Encoding.ASCII.GetBytes(data.ToCharArray());
			ns.Write(outbytes,0,outbytes.Length);
			StreamReader sr = new StreamReader(ns);
			string resp = sr.ReadLine();
			if (!resp.StartsWith("+OK")) 
			{
				log.write("POPListener.GetMessages " + resp);
				return;
			}
			string[] tokens = resp.Split(blankSep);
			int msgNb = Convert.ToInt32(tokens[1]);
			for (int i = 1; i < (msgNb + 1); ++i)
				ProcessMessage(i);
		}

		/// <summary>
		/// Retrieve and parse a message from the POP server.
		/// Calls QueueMessage to unserialize and queue the message.
		/// </summary>
		/// <param name="msgnb">Message number</param>
		private void ProcessMessage(int msgnb) 
		{
			string data = "RETR " + msgnb + "\r\n";
			Byte[] outbytes = System.Text.Encoding.ASCII.GetBytes(data.ToCharArray());
			ns.Write(outbytes,0,outbytes.Length);
			StreamReader sr = new StreamReader(ns);
			string line = sr.ReadLine();
			if (!line.StartsWith("+OK")) 
			{
				log.write("POPListener.GetMessages " + line);
				return;
			}
			// if true, in body part
			bool inBody = false;
			// Mail origin
			string mailFrom = null;
			// Partner message
			StringBuilder body = new StringBuilder("");
			while(true) 
			{
				line = sr.ReadLine();
				if (line.Equals("."))
					break;
				if (inBody) 
				{
					body.Append(line + "\r\n");
					continue;
				}
				if (line.StartsWith("From:")) 
				{
					string[] tokens = line.Split(blankSep);
					for (int i = 0; i < tokens.Length; ++i) 
					{
						if (tokens[i].IndexOf('@') == -1)
							continue;
						if (!tokens[i].StartsWith("<"))
							mailFrom = tokens[i];
						else 
							mailFrom = tokens[i].Substring(1, tokens[i].Length -2);
					}
				}
				if (line.StartsWith("Subject:")) 
				{
					if (line.IndexOf("PageBoxGrid") == -1)
						break;
				}
				if (line.Length == 0)
					inBody = true;
			}
			data = "DELE " + msgnb + "\r\n";
			outbytes = System.Text.Encoding.ASCII.GetBytes(data.ToCharArray());
			ns.Write(outbytes,0,outbytes.Length);
			sr.ReadLine();
			if (body.Length > 0)
				QueueMessage(body, mailFrom);
		}
		/// <summary>
		/// Disconnect from the POP server
		/// </summary>
		private void Disconnect()
		{
			string data = "QUIT" + "\r\n";
			Byte[] outbytes = System.Text.Encoding.ASCII.GetBytes(data.ToCharArray());
			ns.Write(outbytes,0,outbytes.Length);
			ns.Close();
		}

		/// <summary>
		/// Unserialize and queue a message.
		/// </summary>
		/// <param name="body">Message</param>
		/// <param name="mailFrom">Mail origin</param>
		private void QueueMessage(StringBuilder body, string mailFrom) 
		{
			BinaryFormatter bf = new BinaryFormatter();
			ASCIIEncoding ae = new ASCIIEncoding();
			MemoryStream ms = new MemoryStream(ae.GetBytes(body.ToString()));
			DatagramWrapper[] dw = (DatagramWrapper[])bf.Deserialize(ms);
			for (int i = 0; i < dw.Length; ++i) 
			{
				if (dw[i].o is GridCommand) 
				{
					GridCommand gc = (GridCommand)dw[i].o;
					IDictionaryEnumerator ide = clients.GetEnumerator();
					while(ide.MoveNext()) 
					{
						GridQueue gq = (GridQueue)ide.Value;
						gc.Update(gq.client);
					}
					return;
				}
				lock(this) 
				{
					if (dw[i].gridNb != -2) 
					{
						GridQueue gq = (GridQueue)clients[dw[i].gridNb];
						gq.Enqueue(dw[i].o, mailFrom, dw[i].gridNbFrom);
					} 
					else 
					{
						IDictionaryEnumerator ide = clients.GetEnumerator();
						while(ide.MoveNext()) 
						{
							GridQueue gq = (GridQueue)ide.Value;
							gq.Enqueue(dw[i].o, mailFrom, dw[i].gridNbFrom);
						}
					}
				}
			}
		}

		/// <summary>
		/// Called from Grid.SMTPReceive. Calls GridQueue.Dequeue.
		/// </summary>
		/// <param name="gridNb">Grid number of the destination</param>
		/// <returns>Message wrapped with its origin</returns>
		public SMTPMessage Receive(int gridNb) 
		{
			lock(this) 
			{
				GridQueue gq = (GridQueue)clients[gridNb];
				return gq.Dequeue();
			}
		}

		/// <summary>
		/// Called from Grid.SMTPGetMessage. Unsets GridQueue's receivePending.
		/// </summary>
		/// <param name="gridNb">Grid number of the destination</param>
		public void UnsetReceivePending(int gridNb) 
		{
			lock(this) 
			{
				GridQueue gq = (GridQueue)clients[gridNb];
				gq.UnsetReceivePending();
			}
		}
	}

	/// <summary>
	/// 1. Queue responses
	/// 2. Wake up pending receive if needed
	/// </summary>
	public class GridQueue 
	{
		/// <summary>
		/// Response queue
		/// </summary>
		private Queue clientQueue = new Queue();
		/// <summary>
		/// Grid client
		/// </summary>
		public Grid client;
		/// <summary>
		/// If true a receive is pending
		/// </summary>
		private bool receivePending = false;

		/// <summary>
		/// Constructor. Sets the client.
		/// </summary>
		/// <param name="grid">Grid client</param>
		public GridQueue(Grid grid) 
		{
			client = grid;
		}

		/// <summary>
		/// Enqueue an object and wakes up the client if it is waiting.
		/// Version 0.0.5d: calls subscribed message handler.
		/// </summary>
		/// <param name="o">Object to queue</param>
		public void Enqueue(object o, string mailFrom, int gridNbFrom) 
		{
			lock(client) 
			{
				if (client.gc != null) 
					client.gc.Notify(Transport.SMTP,
						client.GetOriginFromMail(mailFrom, gridNbFrom), o);
				else 
				{
					clientQueue.Enqueue(new SMTPMessage(o, mailFrom, gridNbFrom));
					if (receivePending) 
						Monitor.Pulse(client);
				}
			}
		}

		/// <summary>
		/// Retrieve first message in queue.
		/// </summary>
		/// <returns>First message in queue</returns>
		public SMTPMessage Dequeue() 
		{
			lock(client) 
			{
				if (clientQueue.Count == 0) 
				{
					receivePending = true;
					return null;
				}
				receivePending = false;
				return (SMTPMessage)clientQueue.Dequeue();
			}
		}

		/// <summary>
		/// Unset GridQueue's receivePending
		/// </summary>
		public void UnsetReceivePending() 
		{
			lock(client) 
			{
				receivePending = false;
			}
		}

		/// <summary>
		/// Not used for the moment. Property returning the number of messages in queue.
		/// </summary>
		public int MsgQueued 
		{
			get
			{
				lock(client) 
				{
					return clientQueue.Count;
				}
			}
		}
	}

	/// <summary>
	/// Message queued on clientQueue (GridQueue)
	/// </summary>
	public class SMTPMessage 
	{
		/// <summary>
		/// Payload object
		/// </summary>
		public object o;
		/// <summary>
		/// Mail address from the origin
		/// </summary>
		public string mailAddress;
		/// <summary>
		/// Origin grid number
		/// </summary>
		public int gridNb;
		/// <summary>
		/// Constructor. Sets the member variables.
		/// </summary>
		/// <param name="o">Payload object</param>
		/// <param name="mailAddress">Mail address from the origin</param>
		/// <param name="gridNb">Origin grid number</param>
		public SMTPMessage(object o, string mailAddress, int gridNb) 
		{
			this.o = o;
			this.mailAddress = mailAddress;
			this.gridNb = gridNb;
		}
	}
}
