using System;
using System.Collections;
using System.Web;

namespace PageBoxGrid
{
	/// <summary>
	/// <p>Part of the Coordinator set. Transaction management.</p>
	/// <p>Copyright (c) 2002 Alexis Grandemange<br/>
	/// Mail: alexis.grandemange@pagebox.net</p>
	/// <pre>This program is free software; you can redistribute it and/or
	/// modify it under the terms of the GNU Lesser General Public
	/// License as published by the Free Software Foundation; version
	/// 2.1 of the License.
	/// This library is distributed in the hope that it will be useful,
	/// but WITHOUT ANY WARRANTY; without even the implied warranty of
	/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	/// GNU Lesser General Public License for more details.
	/// A copy of the GNU Lesser General Public License lesser.txt should be
	/// included in the distribution.</pre>
	/// </summary>
	public class TransactionCoordinator
	{
		/// <summary>
		/// <p>Key: Web server session ID</p>
		/// <p>Value: Hashtable<ul>
		/// <li>Key: ID of the Transaction commit</li>
		/// <li>Value: Hashtable<ul>
		/// <li>Key: ID of the Transaction participant</li>
		/// <li>Value: TranSub object</li>
		/// </ul></li>
		/// </ul>
		/// </p>
		/// </summary>
		private static Hashtable sessionTrans = new Hashtable();

		/// <summary>
		/// 
		/// </summary>
		/// <param name="sourceID">ID of the Event source</param>
		/// <param name="sessionID">Web server session ID</param>
		/// <param name="ID">ID of the Event destination</param>
		/// <param name="trans">Object provided by the subscriber</param>
		/// <param name="timeout">Time between two Prepare invocations by the TransactionCoordinator</param>
		public static void Subscribe(string sourceID, string sessionID, string ID,
			Transaction trans, int timeout) 
		{
			lock(typeof(TransactionCoordinator)) 
			{
				Hashtable sourceEvents, destEvents;
				bool newDest = false, newSource = false;
				if (!sessionTrans.Contains(sessionID)) 
				{
					sourceEvents = new Hashtable();
					newSource = true;
				}
				else
					sourceEvents = (Hashtable)sessionTrans[sessionID];
				if (!sourceEvents.Contains(sourceID)) 
				{
					destEvents = new Hashtable();
					newDest = true;
				}
				else
					destEvents = (Hashtable)sourceEvents[sourceID];
				TranSub ts = new TranSub(trans, timeout);
				if (!destEvents.Contains(ID))
					destEvents.Add(ID, ts);
				else
					destEvents[ID] = ts;
				if (newDest)
					sourceEvents.Add(sourceID, destEvents);
				if (newSource)
					sessionTrans.Add(sessionID, sourceEvents);
			}
		}

		/// <summary>
		/// Called by the source to Commit the Transaction.
		/// Notes:<ol>
		/// <li>If there is no subscriber the answer is true</li>
		/// <li>An exception = a false answer</li>
		/// <li>If a Prepare returns false, doesn't call Prepare for the remaining subscribers</li>
		/// <li>If a Commit returns false, calls Commit for the remaining subscribers</li>
		/// </ol>
		/// </summary>
		/// <param name="sourceID">ID of the Event source</param>
		/// <param name="sessionID">Web server session ID</param>
		/// <param name="req">Object to pass to the Prepare and Commit methods of the destination.
		/// </param>
		/// <param name="resp">Hashtable of responses whose key is the destination ID.
		/// The Hashtable contains either Prepare responses (Prepare failure) or Commit responses</param>
		/// <param name="request">HTTP request</param>
		/// <returns>True if all Subscriber Prepare and Commit have succeeded</returns>
		public static bool Commit(string sourceID, string sessionID, object req, ref Hashtable resp,
			HttpRequest request) 
		{
			lock(typeof(TransactionCoordinator)) 
			{
				if (!sessionTrans.Contains(sessionID))
					return true;
				Hashtable sourceEvents = (Hashtable)sessionTrans[sessionID];
				if (!sourceEvents.Contains(sourceID))
					return true;
				Hashtable destEvents = (Hashtable)sourceEvents[sourceID];
				IDictionaryEnumerator ide = destEvents.GetEnumerator();
				bool rc = true;
				while(ide.MoveNext()) 
				{
					TranSub ts = (TranSub)ide.Value;
					if ((ts.timeout == 0) || (ts.dt == DateTime.MinValue) || (ts.dt > DateTime.Now)) 
					{
						try 
						{
							object r = null;
							if (!ts.trans.Prepare(sourceID, req, ref r, request)) 
							{
								rc = false;
								if ((r != null) && (resp != null))
									resp[ide.Key] = r;
								break;
							}
						}
						catch(Exception e) 
						{
							if (resp != null)
								resp[ide.Key] = e;
							rc = false;
							break;
						}
						if (ts.timeout == 0)
							continue;
						ts.dt = DateTime.Now;
						ts.dt.AddSeconds(ts.timeout);
					}
				}
				if (!rc)
					return false;
				ide.Reset();
				while(ide.MoveNext()) 
				{
					TranSub ts = (TranSub)ide.Value;
					try 
					{
						object r = null;
						if (!ts.trans.Commit(sourceID, req, ref r, request)) 
							rc = false;
						if ((r != null) && (resp != null))
							resp[ide.Key] = r;
					}
					catch(Exception e) 
					{
						if (resp != null)
							resp[ide.Key] = e;
						rc = false;
					}
				}
				return rc;
			}
		}

		/// <summary>
		/// Called by the source to Rollback the Transaction.
		/// Notes:<ol>
		/// <li>If there is no subscriber the answer is true</li>
		/// <li>An exception = a false answer</li>
		/// <li>All subscribers are notified even if a subscriber answers false</li>
		/// </ol>
		/// </summary>
		/// <param name="sourceID">ID of the Event source</param>
		/// <param name="sessionID">Web server session ID</param>
		/// <param name="req">Object to pass to the Rollback method of the destination.</param>
		/// <param name="resp">Hashtable of responses whose key is the destination ID.</param>
		/// <param name="request">HTTP request</param>
		/// <returns>True if all Subscriber Rollback have succeeded</returns>
		public static bool Rollback(string sourceID, string sessionID, object req, ref Hashtable resp,
			HttpRequest request) 
		{
			lock(typeof(TransactionCoordinator)) 
			{
				if (!sessionTrans.Contains(sessionID))
					return true;
				Hashtable sourceEvents = (Hashtable)sessionTrans[sessionID];
				if (!sourceEvents.Contains(sourceID))
					return true;
				Hashtable destEvents = (Hashtable)sourceEvents[sourceID];
				IDictionaryEnumerator ide = destEvents.GetEnumerator();
				bool rc = true;
				while(ide.MoveNext()) 
				{
					TranSub ts = (TranSub)ide.Value;
					try 
					{
						object r = null;
						if (!ts.trans.Rollback(sourceID, req, ref r, request)) 
							rc = false;
						if ((r != null) && (resp != null))
							resp[ide.Key] = r;
					}
					catch(Exception e) 
					{
						if (resp != null)
							resp[ide.Key] = e;
						rc = false;
					}
				}
				return rc;
			}
		}
	}

	/// <summary>
	/// Represents a Transaction subscriber (participant)
	/// </summary>
	class TranSub 
	{
		/// <summary>
		/// Time between two Prepare invocations by the TransactionCoordinator
		/// </summary>
		public int timeout;
		/// <summary>
		/// Object provided by the subscriber
		/// </summary>
		public Transaction trans;
		public DateTime dt = DateTime.MinValue;

		/// <summary>
		/// Constructor. Set member variables.
		/// </summary>
		/// <param name="trans">Object provided by the subscriber</param>
		/// <param name="timeout">Time between two Prepare invocations by the TransactionCoordinator</param>
		public TranSub(Transaction trans, int timeout) 
		{
			this.trans = trans;
			this.timeout = timeout;
		}
	}

	/// <summary>
	/// Interface provided by the subscriber
	/// </summary>
	public interface Transaction 
	{
		/// <summary>
		/// Ask to the subscriber if it is ready to commit the transaction (2-phase commit)
		/// </summary>
		/// <param name="sourceID">ID of the Event source</param>
		/// <param name="req">Object passed to the Prepare method</param>
		/// <param name="resp">Object returned by the Prepare method</param>
		/// <param name="request">HTTP request object of the source</param>
		/// <returns>Return false to indicate a failure</returns>
		bool Prepare(string sourceID, object req, ref object resp, HttpRequest request);

		/// <summary>
		/// Ask to the subscriber to commit its part of the transaction (2-phase commit)
		/// </summary>
		/// <param name="sourceID">ID of the Event source</param>
		/// <param name="req">Object passed to the Commit method</param>
		/// <param name="resp">Object returned by the Commit method</param>
		/// <param name="request">HTTP request object of the source</param>
		/// <returns>Return false to indicate a failure</returns>
		bool Commit(string sourceID, object req, ref object resp, HttpRequest request);

		/// <summary>
		/// Ask to the subscriber to rollback its part of the transaction (2-phase commit)
		/// </summary>
		/// <param name="sourceID">ID of the Event source</param>
		/// <param name="req">Object passed to the Rollback method</param>
		/// <param name="resp">Object returned by the Rollback method</param>
		/// <param name="request">HTTP request object of the source</param>
		/// <returns>Return false to indicate a failure</returns>
		bool Rollback(string sourceID, object req, ref object resp, HttpRequest request);
	}
}
