using System;
using System.Net;
using System.Collections;
using System.Threading;

namespace PageBoxGrid
{
	/// <summary>
	/// <p>Queue used for UDP unicast and multicast.
	/// See <a href="http://pagebox.net/grid.html">Grid</a> for explanations.</p>
	/// <p>Copyright (c) 2002 Alexis Grandemange<br/>
	/// Mail: alexis.grandemange@pagebox.net</p>
	/// <pre>This program is free software; you can redistribute it and/or
	/// modify it under the terms of the GNU Lesser General Public
	/// License as published by the Free Software Foundation; version
	/// 2.1 of the License.
	/// This library is distributed in the hope that it will be useful,
	/// but WITHOUT ANY WARRANTY; without even the implied warranty of
	/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	/// GNU Lesser General Public License for more details.
	/// A copy of the GNU Lesser General Public License lesser.txt should be
	/// included in the distribution.</pre>
	/// </summary>
	public class UDPQueue
	{
		/// <summary>
		/// Response queue
		/// </summary>
		private Queue clientQueue = new Queue();
		/// <summary>
		/// Grid client
		/// </summary>
		public Grid client;
		/// <summary>
		/// If true a receive is pending
		/// </summary>
		private bool receivePending = false;

		/// <summary>
		/// Constructor. Sets the client.
		/// </summary>
		/// <param name="grid">Grid client</param>
		public UDPQueue(Grid grid) 
		{
			client = grid;
		}

		/// <summary>
		/// Enqueue an object and wakes up the client if it is waiting.
		/// Checks if destination == origin (happens in multicast).
		/// Version 0.0.5d: calls the message handler if subscribed.
		/// </summary>
		/// <param name="o">Object to queue</param>
		public void Enqueue(object o, IPAddress addrFrom, int portFrom, int gridNbFrom) 
		{
			if ((client.ge.addr == addrFrom) && (client.ge.UDPPort == portFrom) &&
				(client.ge.gridNb == gridNbFrom))
				return;
			lock(client) 
			{
				if (client.gc != null)
					client.gc.Notify(Transport.UDP, client.GetOriginFromUDP(addrFrom, portFrom,
						gridNbFrom), o);
				else 
				{
					clientQueue.Enqueue(new UDPMessage(o, addrFrom, portFrom, gridNbFrom));
					if (receivePending) 
						Monitor.Pulse(client);
				}
			}
		}

		/// <summary>
		/// Retrieve first message in queue.
		/// </summary>
		/// <returns>First message in queue</returns>
		public UDPMessage Dequeue() 
		{
			lock(client) 
			{
				if (clientQueue.Count == 0) 
				{
					receivePending = true;
					return null;
				}
				receivePending = false;
				return (UDPMessage)clientQueue.Dequeue();
			}
		}

		/// <summary>
		/// Unset GridQueue's receivePending
		/// </summary>
		public void UnsetReceivePending() 
		{
			lock(client) 
			{
				receivePending = false;
			}
		}

		/// <summary>
		/// Not used for the moment. Property returning the number of messages in queue.
		/// </summary>
		public int MsgQueued 
		{
			get
			{
				lock(client) 
				{
					return clientQueue.Count;
				}
			}
		}
	}

	/// <summary>
	/// Message queued on clientQueue (UDPQueue)
	/// </summary>
	public class UDPMessage 
	{
		/// <summary>
		/// Payload object
		/// </summary>
		public object o;
		/// <summary>
		/// IP address of the origin
		/// </summary>
		public IPAddress addr;
		/// <summary>
		/// UDP port of the origin
		/// </summary>
		public int port;
		/// <summary>
		/// Origin grid number
		/// </summary>
		public int gridNb;
		/// <summary>
		/// Constructor. Sets the member variables.
		/// </summary>
		/// <param name="o">Payload object</param>
		/// <param name="addr">IP address of the origin</param>
		/// <param name="port">UDP port of the origin</param>
		/// <param name="gridNb">Origin grid number</param>
		public UDPMessage(object o, IPAddress addr, int port, int gridNb) 
		{
			this.o = o;
			this.addr = addr;
			this.port = port;
			this.gridNb = gridNb;
		}
	}

	/// <summary>
	/// Message serialized in UDP and SMTP.
	/// </summary>
	public class DatagramWrapper 
	{
		/// <summary>
		/// Payload
		/// </summary>
		public object o = null;
		/// <summary>
		/// Grid number of the origin
		/// </summary>
		public int gridNbFrom = -1;
		/// <summary>
		/// Grid number of the destination
		/// </summary>
		public int gridNb = -1;
		/// <summary>
		/// Class of the payload
		/// </summary>
		public string objectClass = null;

		/// <summary>
		/// Constructor. Set member variables.
		/// </summary>
		/// <param name="o">Payload</param>
		/// <param name="gridNbFrom">Grid number of the origin</param>
		/// <param name="gridNb">Grid number of the destination</param>
		public DatagramWrapper(object o, int gridNbFrom, int gridNb) 
		{
			this.o = o;
			this.gridNb = gridNb;
			this.gridNbFrom = gridNbFrom;
			objectClass = o.GetType().FullName;
		}

		/// <summary>
		/// Default constructor.
		/// </summary>
		public DatagramWrapper() 
		{
		}
	}

	/// <summary>
	/// Message serialized in multicast mode
	/// </summary>
	public class MulticastWrapper 
	{
		/// <summary>
		/// Payload
		/// </summary>
		public object o = null;
		/// <summary>
		/// Grid number of the origin
		/// </summary>
		public int gridNbFrom = -1;
		/// <summary>
		/// Class of the payload
		/// </summary>
		public string objectClass = null;
		/// <summary>
		/// IP address of the origin
		/// </summary>
		public IPAddress addrFrom;
		/// <summary>
		/// Port of the origin
		/// </summary>
		public int portFrom;
		/// <summary>
		/// Constructor. Set member variables.
		/// </summary>
		/// <param name="o">Payload</param>
		/// <param name="gridNbFrom">Grid number of the origin</param>
		/// <param name="addrFrom">IP address of the origin</param>
		/// <param name="portFrom">Port of the origin</param>
		public MulticastWrapper(object o, int gridNbFrom, IPAddress addrFrom, int portFrom) 
		{
			this.o = o;
			this.gridNbFrom = gridNbFrom;
			objectClass = o.GetType().FullName;
			this.addrFrom = addrFrom;
			this.portFrom = portFrom;
		}
	}
}
